@@ -2,6 +2,7 @@
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
  *
  * Copyright (C) 2000-2005 by Alfons Hoogervorst & The Sylpheed Claws Team.
+ * Copyright (C) 1999-2012 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -36,6 +37,7 @@
 #include "utils.h"
 #include "addressbook.h"
 #include "main.h"
+#include "prefs_common.h"
 
 /* How it works:
  *
@@ -92,6 +94,11 @@ static gchar	   *completion_prefix;		/* last prefix. (this is cached here
 
 /*******************************************************************************/
 
+
+static void address_completion_entry_changed		(GtkEditable *editable,
+							 gpointer     data);
+
+
 /* completion_func() - used by GTK to find the string data to be used for 
  * completion 
  */
@@ -154,24 +161,34 @@ static gint address_entry_find_func(gconstpointer a, gconstpointer b)
 	return 0;
 }
 
+static void add_completion_entry(const gchar *str, address_entry *ae)
+{
+	completion_entry *ce;
+
+	if (!str || *str == '\0')
+		return;
+	if (!ae)
+		return;
+
+	ce = g_new0(completion_entry, 1);
+	/* GCompletion list is case sensitive */
+	ce->string = g_utf8_strdown(str, -1);
+	ce->ref = ae;
+	completion_list = g_list_append(completion_list, ce);
+}
+
 /* add_address() - adds address to the completion list. this function looks
  * complicated, but it's only allocation checks.
  */
-static gint add_address(const gchar *name, const gchar *address, const gchar *nickname)
+static gint add_address(const gchar *name, const gchar *firstname, const gchar *lastname, const gchar *nickname, const gchar *address)
 {
-	address_entry    *ae;
-	completion_entry *name_ce = NULL;
-	completion_entry *nick_ce = NULL;
-	completion_entry *addr_ce;
-	GList            *found;
+	address_entry *ae;
+	GList         *found;
 
-	if (!address) return -1;
+	if (!address || *address == '\0')
+		return -1;
 
-	if (name && *name)
-		name_ce = g_new0(completion_entry, 1);
-	if (nickname && *nickname)
-		nick_ce = g_new0(completion_entry, 1);
-	addr_ce = g_new0(completion_entry, 1);
+	/* debugg_print("add_address: [%s] [%s] [%s] [%s] [%s]\n", name, firstname, lastname, nickname, address); */
 
 	ae = g_new0(address_entry, 1);
 	ae->name    = g_strdup(name ? name : "");
@@ -184,25 +201,22 @@ static gint add_address(const gchar *name, const gchar *address, const gchar *ni
 		ae = (address_entry *)found->data;
 	} else
 		address_list = g_list_append(address_list, ae);
+ 
+	if (name) {
+		const gchar *p = name;
 
-	/* GCompletion list is case sensitive */
-	if (name_ce)
-		name_ce->string = g_utf8_strdown(name, -1);
-	if (nick_ce)
-		nick_ce->string = g_utf8_strdown(nickname, -1);
-	addr_ce->string = g_utf8_strdown(address, -1);
-
-	if (name_ce)
-		name_ce->ref = ae;
-	if (nick_ce)
-		nick_ce->ref = ae;
-	addr_ce->ref = ae;
-
-	if (name_ce)
-		completion_list = g_list_append(completion_list, name_ce);
-	if (nick_ce)
-		completion_list = g_list_append(completion_list, nick_ce);
-	completion_list = g_list_append(completion_list, addr_ce);
+		while (*p != '\0') {
+			add_completion_entry(p, ae);
+			while (*p && *p != '-' && *p != '.' && !g_ascii_isspace(*p))
+				p++;
+			while (*p == '-' || *p == '.' || g_ascii_isspace(*p))
+				p++;
+		}
+	}
+	add_completion_entry(firstname, ae);
+	add_completion_entry(lastname, ae);
+	add_completion_entry(nickname, ae);
+	add_completion_entry(address, ae);
 
 	return 0;
 }
@@ -210,7 +224,7 @@ static gint add_address(const gchar *name, const gchar *address, const gchar *ni
 /* read_address_book()
  */ 
 static void read_address_book(void) {	
-	addressbook_load_completion( add_address );
+	addressbook_load_completion_full( add_address );
 }
 
 /* start_address_completion() - returns the number of addresses 
@@ -291,12 +305,27 @@ gchar *get_address_from_edit(GtkEntry *entry, gint *start_pos)
 void replace_address_in_edit(GtkEntry *entry, const gchar *newtext,
 			     gint start_pos)
 {
+	gchar *origtext;
+
 	if (!newtext) return;
 
+	debug_print("replace_address_in_edit: %s\n", newtext);
+
+	origtext = gtk_editable_get_chars(GTK_EDITABLE(entry), start_pos, -1);
+	if (!strcmp(origtext, newtext)) {
+		g_free(origtext);
+		return;
+	}
+	g_free(origtext);
+
+	g_signal_handlers_block_by_func
+		(entry, address_completion_entry_changed, NULL);
 	gtk_editable_delete_text(GTK_EDITABLE(entry), start_pos, -1);
 	gtk_editable_insert_text(GTK_EDITABLE(entry), newtext, strlen(newtext),
 				 &start_pos);
 	gtk_editable_set_position(GTK_EDITABLE(entry), -1);
+	g_signal_handlers_unblock_by_func
+		(entry, address_completion_entry_changed, NULL);
 }
 
 #if 0
@@ -528,21 +557,22 @@ static gboolean address_completion_entry_key_pressed	(GtkEntry    *entry,
 static gboolean address_completion_complete_address_in_entry
 							(GtkEntry    *entry,
 							 gboolean     next);
-static void address_completion_create_completion_window	(GtkEntry    *entry);
+static void address_completion_create_completion_window	(GtkEntry    *entry,
+							 gboolean     select_next);
 
 static void completion_window_select_row(GtkCList	 *clist,
 					 gint		  row,
 					 gint		  col,
 					 GdkEvent	 *event,
-					 GtkWidget	**completion_window);
+					 GtkWidget	**window);
 static gboolean completion_window_button_press
 					(GtkWidget	 *widget,
 					 GdkEventButton  *event,
-					 GtkWidget	**completion_window);
+					 GtkWidget	**window);
 static gboolean completion_window_key_press
 					(GtkWidget	 *widget,
 					 GdkEventKey	 *event,
-					 GtkWidget	**completion_window);
+					 GtkWidget	**window);
 
 
 static void completion_window_advance_to_row(GtkCList *clist, gint row)
@@ -669,6 +699,9 @@ void address_completion_register_entry(GtkEntry *entry)
 			(G_CALLBACK(address_completion_entry_key_pressed),
 			 COMPLETION_UNIQUE_DATA, NULL),
 		 FALSE);
+	g_signal_connect(G_OBJECT(entry), "changed",
+			 G_CALLBACK(address_completion_entry_changed),
+			 NULL);
 }
 
 void address_completion_unregister_entry(GtkEntry *entry)
@@ -709,19 +742,23 @@ static void address_completion_mainwindow_set_focus(GtkWindow *window,
 						    gpointer   data)
 {
 	if (widget && GTK_IS_ENTRY(widget) &&
-	    g_object_get_data(G_OBJECT(widget), ENTRY_DATA_TAB_HOOK)) 
+	    g_object_get_data(G_OBJECT(widget), ENTRY_DATA_TAB_HOOK)) {
 		clear_completion_cache();
+	}
 }
 
+static GtkWidget *completion_window;
+
 /* watch for tabs in one of the address entries. if no tab then clear the
  * completion cache */
 static gboolean address_completion_entry_key_pressed(GtkEntry    *entry,
 						     GdkEventKey *ev,
 						     gpointer     data)
 {
-	if (ev->keyval == GDK_Tab) {
+	if (!prefs_common.fullauto_completion_mode && ev->keyval == GDK_Tab &&
+	    !completion_window) {
 		if (address_completion_complete_address_in_entry(entry, TRUE)) {
-			address_completion_create_completion_window(entry);
+			address_completion_create_completion_window(entry, TRUE);
 			/* route a void character to the default handler */
 			/* this is a dirty hack; we're actually changing a key
 			 * reported by the system. */
@@ -729,6 +766,21 @@ static gboolean address_completion_entry_key_pressed(GtkEntry    *entry,
 			ev->state &= ~GDK_SHIFT_MASK;
 			return TRUE;
 		}
+	}
+
+	if (!completion_window)
+		return FALSE;
+
+	if (       ev->keyval == GDK_Up
+		|| ev->keyval == GDK_Down
+		|| ev->keyval == GDK_Page_Up
+		|| ev->keyval == GDK_Page_Down
+		|| ev->keyval == GDK_Return
+		|| ev->keyval == GDK_Escape
+		|| ev->keyval == GDK_Tab
+		|| ev->keyval == GDK_ISO_Left_Tab) {
+		completion_window_key_press(completion_window, ev, &completion_window);
+		return TRUE;
 	} else if (ev->keyval == GDK_Shift_L
 		|| ev->keyval == GDK_Shift_R
 		|| ev->keyval == GDK_Control_L
@@ -740,12 +792,38 @@ static gboolean address_completion_entry_key_pressed(GtkEntry    *entry,
 		|| ev->keyval == GDK_Alt_L
 		|| ev->keyval == GDK_Alt_R) {
 		/* these buttons should not clear the cache... */
-	} else
+	} else {
 		clear_completion_cache();
+		gtk_widget_destroy(completion_window);
+		completion_window = NULL;
+	}
 
 	return FALSE;
 }
 
+static void address_completion_entry_changed(GtkEditable *editable,
+					     gpointer data)
+{
+	GtkEntry *entry = GTK_ENTRY(editable);
+
+	if (!prefs_common.fullauto_completion_mode)
+		return;
+
+	g_signal_handlers_block_by_func
+		(editable, address_completion_entry_changed, data);
+	if (address_completion_complete_address_in_entry(entry, TRUE)) {
+		address_completion_create_completion_window(entry, FALSE);
+	} else {
+		clear_completion_cache();
+		if (completion_window) {
+			gtk_widget_destroy(completion_window);
+			completion_window = NULL;
+		}
+	}
+	g_signal_handlers_unblock_by_func
+		(editable, address_completion_entry_changed, data);
+}
+
 /* initialize the completion cache and put first completed string
  * in entry. this function used to do back cycling but this is not
  * currently used. since the address completion behaviour has been
@@ -792,15 +870,17 @@ static gboolean address_completion_complete_address_in_entry(GtkEntry *entry,
 	return completed;
 }
 
-static void address_completion_create_completion_window(GtkEntry *entry_)
+static void address_completion_create_completion_window(GtkEntry *entry_,
+							gboolean select_next)
 {
-	static GtkWidget *completion_window;
 	gint x, y, height, width, depth;
 	GtkWidget *scroll, *clist;
 	GtkRequisition r;
 	guint count = 0;
 	GtkWidget *entry = GTK_WIDGET(entry_);
 
+	debug_print("address_completion_create_completion_window\n");
+
 	if (completion_window) {
 		gtk_widget_destroy(completion_window);
 		completion_window = NULL;
@@ -870,7 +950,9 @@ static void address_completion_create_completion_window(GtkEntry *entry_)
 	/* this gets rid of the irritating focus rectangle that doesn't
 	 * follow the selection */
 	GTK_WIDGET_UNSET_FLAGS(clist, GTK_CAN_FOCUS);
-	gtk_clist_select_row(GTK_CLIST(clist), 1, 0);
+	gtk_clist_select_row(GTK_CLIST(clist), select_next ? 1 : 0, 0);
+
+	debug_print("address_completion_create_completion_window done\n");
 }
 
 
@@ -878,14 +960,14 @@ static void address_completion_create_completion_window(GtkEntry *entry_)
  * note: event is NULL if selected by anything else than a mouse button. */
 static void completion_window_select_row(GtkCList *clist, gint row, gint col,
 					 GdkEvent *event,
-					 GtkWidget **completion_window)
+					 GtkWidget **window)
 {
 	GtkEntry *entry;
 
-	g_return_if_fail(completion_window != NULL);
-	g_return_if_fail(*completion_window != NULL);
+	g_return_if_fail(window != NULL);
+	g_return_if_fail(*window != NULL);
 
-	entry = GTK_ENTRY(g_object_get_data(G_OBJECT(*completion_window),
+	entry = GTK_ENTRY(g_object_get_data(G_OBJECT(*window),
 					    WINDOW_DATA_COMPL_ENTRY));
 	g_return_if_fail(entry != NULL);
 
@@ -895,8 +977,8 @@ static void completion_window_select_row(GtkCList *clist, gint row, gint col,
 		return;
 
 	clear_completion_cache();
-	gtk_widget_destroy(*completion_window);
-	*completion_window = NULL;
+	gtk_widget_destroy(*window);
+	*window = NULL;
 }
 
 /* completion_window_button_press() - check is mouse click is anywhere
@@ -904,17 +986,17 @@ static void completion_window_select_row(GtkCList *clist, gint row, gint col,
  * window is destroyed, and the original prefix is restored */
 static gboolean completion_window_button_press(GtkWidget *widget,
 					       GdkEventButton *event,
-					       GtkWidget **completion_window)
+					       GtkWidget **window)
 {
 	GtkWidget *event_widget, *entry;
 	gchar *prefix;
 	gint cursor_pos;
 	gboolean restore = TRUE;
 
-	g_return_val_if_fail(completion_window != NULL, FALSE);
-	g_return_val_if_fail(*completion_window != NULL, FALSE);
+	g_return_val_if_fail(window != NULL, FALSE);
+	g_return_val_if_fail(*window != NULL, FALSE);
 
-	entry = GTK_WIDGET(g_object_get_data(G_OBJECT(*completion_window),
+	entry = GTK_WIDGET(g_object_get_data(G_OBJECT(*window),
 					     WINDOW_DATA_COMPL_ENTRY));
 	g_return_val_if_fail(entry != NULL, FALSE);
 
@@ -939,28 +1021,30 @@ static gboolean completion_window_button_press(GtkWidget *widget,
 	}
 
 	clear_completion_cache();
-	gtk_widget_destroy(*completion_window);
-	*completion_window = NULL;
+	gtk_widget_destroy(*window);
+	*window = NULL;
 
 	return TRUE;
 }
 
 static gboolean completion_window_key_press(GtkWidget *widget,
 					    GdkEventKey *event,
-					    GtkWidget **completion_window)
+					    GtkWidget **window)
 {
-	GdkEventKey tmp_event;
 	GtkWidget *entry;
 	gchar *prefix;
 	gint cursor_pos;
 	GtkWidget *clist;
 
-	g_return_val_if_fail(completion_window != NULL, FALSE);
-	g_return_val_if_fail(*completion_window != NULL, FALSE);
+	g_return_val_if_fail(window != NULL, FALSE);
+	g_return_val_if_fail(*window != NULL, FALSE);
 
-	entry = GTK_WIDGET(g_object_get_data(G_OBJECT(*completion_window),
+	if (!is_completion_pending())
+		g_warning("completion is not pending!\n");
+
+	entry = GTK_WIDGET(g_object_get_data(G_OBJECT(*window),
 					     WINDOW_DATA_COMPL_ENTRY));
-	clist = GTK_WIDGET(g_object_get_data(G_OBJECT(*completion_window),
+	clist = GTK_WIDGET(g_object_get_data(G_OBJECT(*window),
 					     WINDOW_DATA_COMPL_CLIST));
 	g_return_val_if_fail(entry != NULL, FALSE);
 
@@ -988,15 +1072,18 @@ static gboolean completion_window_key_press(GtkWidget *widget,
 	}
 
 	/* look for presses that accept the selection */
-	if (event->keyval == GDK_Return || event->keyval == GDK_space) {
+	if (event->keyval == GDK_Return ||
+	    (!prefs_common.fullauto_completion_mode &&
+	     event->keyval == GDK_space)) {
 		/* insert address only if shift or control is pressed */
-		if (event->state & (GDK_SHIFT_MASK|GDK_CONTROL_MASK)) {
+		if (event->state & (GDK_SHIFT_MASK|GDK_CONTROL_MASK) ||
+		    prefs_common.always_add_address_only) {
 			completion_window_apply_selection_address_only
 				(GTK_CLIST(clist), GTK_ENTRY(entry));
 		}
 		clear_completion_cache();
-		gtk_widget_destroy(*completion_window);
-		*completion_window = NULL;
+		gtk_widget_destroy(*window);
+		*window = NULL;
 		return FALSE;
 	}
 
@@ -1015,28 +1102,35 @@ static gboolean completion_window_key_press(GtkWidget *widget,
 	}
 
 	/* other key, let's restore the prefix (orignal text) */
-	prefix = get_complete_address(0);
-	g_free(get_address_from_edit(GTK_ENTRY(entry), &cursor_pos));
-	replace_address_in_edit(GTK_ENTRY(entry), prefix, cursor_pos);
-	g_free(prefix);
-	clear_completion_cache();
+	if (!prefs_common.fullauto_completion_mode ||
+	    event->keyval == GDK_Escape) {
+		prefix = get_complete_address(0);
+		g_free(get_address_from_edit(GTK_ENTRY(entry), &cursor_pos));
+		replace_address_in_edit(GTK_ENTRY(entry), prefix, cursor_pos);
+		g_free(prefix);
+	}
 
 	/* make sure anything we typed comes in the edit box */
-	if (event->length > 0 && event->keyval != GDK_Escape) {
-		tmp_event.type       = event->type;
-		tmp_event.window     = entry->window;
-		tmp_event.send_event = TRUE;
-		tmp_event.time       = event->time;
-		tmp_event.state      = event->state;
-		tmp_event.keyval     = event->keyval;
-		tmp_event.length     = event->length;
-		tmp_event.string     = event->string;
-		gtk_widget_event(entry, (GdkEvent *)&tmp_event);
+	if ((!prefs_common.fullauto_completion_mode && event->length > 0 &&
+	     event->keyval != GDK_Escape) ||
+	    (prefs_common.fullauto_completion_mode &&
+	     event->keyval != GDK_Escape)) {
+		GtkWidget *pwin = entry;
+
+		while ((pwin = gtk_widget_get_parent(pwin)) != NULL) {
+			if (GTK_WIDGET_TOPLEVEL(pwin)) {
+				gtk_window_propagate_key_event
+					(GTK_WINDOW(pwin), event);
+				if (prefs_common.fullauto_completion_mode)
+					return TRUE;
+			}
+		}
 	}
 
 	/* and close the completion window */
-	gtk_widget_destroy(*completion_window);
-	*completion_window = NULL;
+	clear_completion_cache();
+	gtk_widget_destroy(*window);
+	*window = NULL;
 
 	return TRUE;
 }
