@@ -1,6 +1,6 @@
 /*
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
- * Copyright (C) 1999-2007 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2012 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,12 +25,22 @@
 #include <gtk/gtkexpander.h>
 #include <gtk/gtkstock.h>
 
+#ifdef G_OS_WIN32
+#  define COBJMACROS
+#  include <windows.h>
+#  include <objbase.h>
+#  include <objidl.h>
+#  include <shlobj.h>
+#endif
+
 #include "main.h"
 #include "filesel.h"
 #include "manage_window.h"
 #include "alertpanel.h"
 #include "utils.h"
 #include "prefs_common.h"
+#include "inc.h"
+
 
 static GSList *filesel_select_file_full	(const gchar		*title,
 					 const gchar		*file,
@@ -49,6 +59,14 @@ static GtkFileChooserConfirmation filesel_confirm_overwrite_cb
 					(GtkFileChooser		*chooser,
 					 gpointer		 data);
 #endif
+#ifdef G_OS_WIN32
+static gboolean is_ext_lnk		(const gchar		*filename);
+static gchar *filesel_get_link		(const gchar		*link_file);
+static GSList *filesel_resolve_link	(GtkFileChooser		*chooser,
+					 GSList			*list,
+					 gboolean		*dir_selected);
+#endif
+
 
 gchar *filesel_select_file(const gchar *title, const gchar *file,
 			   GtkFileChooserAction action)
@@ -72,27 +90,82 @@ GSList *filesel_select_files(const gchar *title, const gchar *file,
 	return filesel_select_file_full(title, file, action, TRUE);
 }
 
+static void filesel_change_dir_for_action(GtkFileChooserAction action)
+{
+	const gchar *cwd = NULL;
+
+	switch (action) {
+	case GTK_FILE_CHOOSER_ACTION_OPEN:
+		if (prefs_common.prev_open_dir &&
+		    is_dir_exist(prefs_common.prev_open_dir))
+			cwd = prefs_common.prev_open_dir;
+		else {
+			g_free(prefs_common.prev_open_dir);
+			prefs_common.prev_open_dir = NULL;
+		}
+		break;
+	case GTK_FILE_CHOOSER_ACTION_SAVE:
+		if (prefs_common.prev_save_dir &&
+		    is_dir_exist(prefs_common.prev_save_dir))
+			cwd = prefs_common.prev_save_dir;
+		else {
+			g_free(prefs_common.prev_save_dir);
+			prefs_common.prev_save_dir = NULL;
+		}
+		break;
+	case GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER:
+		if (prefs_common.prev_folder_dir &&
+		    is_dir_exist(prefs_common.prev_folder_dir))
+			cwd = prefs_common.prev_folder_dir;
+		else {
+			g_free(prefs_common.prev_folder_dir);
+			prefs_common.prev_folder_dir = NULL;
+		}
+		break;
+	default:
+		break;
+	}
+
+	if (cwd)
+		change_dir(cwd);
+	else
+		change_dir(get_document_dir());
+}
+
+static void filesel_save_dir_for_action(GtkFileChooserAction action,
+					const gchar *cwd)
+{
+	switch (action) {
+	case GTK_FILE_CHOOSER_ACTION_OPEN:
+		g_free(prefs_common.prev_open_dir);
+		prefs_common.prev_open_dir = g_strdup(cwd);
+		break;
+	case GTK_FILE_CHOOSER_ACTION_SAVE:
+		g_free(prefs_common.prev_save_dir);
+		prefs_common.prev_save_dir = g_strdup(cwd);
+		break;
+	case GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER:
+		g_free(prefs_common.prev_folder_dir);
+		prefs_common.prev_folder_dir = g_strdup(cwd);
+		break;
+	default:
+		break;
+	}
+}
+
 static GSList *filesel_select_file_full(const gchar *title, const gchar *file,
 					GtkFileChooserAction action,
 					gboolean multiple)
 {
-	static GHashTable *path_table = NULL;
 	gchar *cwd;
 	GtkWidget *dialog;
 	gchar *prev_dir;
 	static gboolean save_expander_expanded = FALSE;
 	GSList *list = NULL;
 
-	if (!path_table)
-		path_table = g_hash_table_new_full(g_str_hash, g_str_equal,
-						   g_free, g_free);
-
 	prev_dir = g_get_current_dir();
 
-	if ((cwd = g_hash_table_lookup(path_table, title)) != NULL)
-		change_dir(cwd);
-	else
-		change_dir(get_document_dir());
+	filesel_change_dir_for_action(action);
 
 	dialog = filesel_create(title, action);
 
@@ -125,17 +198,38 @@ static GSList *filesel_select_file_full(const gchar *title, const gchar *file,
 	change_dir(prev_dir);
 	g_free(prev_dir);
 
+	inc_lock();
+
+#ifdef G_OS_WIN32
+again:
+#endif
 	if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
 		list = gtk_file_chooser_get_filenames(GTK_FILE_CHOOSER(dialog));
+#ifdef G_OS_WIN32
+		if (list) {
+			/* resolve Windows shortcut */
+			gboolean dir_selected = FALSE;
+			list = filesel_resolve_link(GTK_FILE_CHOOSER(dialog),
+						    list, &dir_selected);
+			if (!list) {
+				if (!dir_selected)
+					alertpanel_error(_("The link target not found."));
+				goto again;
+			}
+		}
+#endif
 		if (list) {
 			cwd = gtk_file_chooser_get_current_folder
 				(GTK_FILE_CHOOSER(dialog));
-			if (cwd)
-				g_hash_table_replace
-					(path_table, g_strdup(title), cwd);
+			if (cwd) {
+				filesel_save_dir_for_action(action, cwd);
+				g_free(cwd);
+			}
 		}
 	}
 
+	inc_unlock();
+
 	if (action == GTK_FILE_CHOOSER_ACTION_SAVE)
 		save_expander_expanded =
 			filesel_save_expander_get_expanded(dialog);
@@ -175,7 +269,7 @@ gchar *filesel_select_dir(const gchar *dir)
 	GSList *list;
 	gchar *selected = NULL;
 
-	list = filesel_select_file_full(_("Select directory"), dir,
+	list = filesel_select_file_full(_("Select folder"), dir,
 					GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
 					FALSE);
 	if (list) {
@@ -271,6 +365,20 @@ static GtkFileChooserConfirmation filesel_confirm_overwrite_cb
 	if (filename && is_file_exist(filename)) {
 		AlertValue aval;
 
+#ifdef G_OS_WIN32
+		gchar *target = NULL;
+
+		if (is_ext_lnk(filename) &&
+		    (target = filesel_get_link(filename)) != NULL &&
+		    is_dir_exist(target)) {
+			gtk_file_chooser_set_current_folder(chooser, target);
+		    	g_free(target);
+		    	g_free(filename);
+		    	return GTK_FILE_CHOOSER_CONFIRMATION_SELECT_AGAIN;
+		}
+		g_free(target);
+#endif
+
 		aval = alertpanel(_("Overwrite existing file"),
 				  _("The file already exists. Do you want to replace it?"),
 				  GTK_STOCK_YES, GTK_STOCK_NO, NULL);
@@ -285,3 +393,90 @@ static GtkFileChooserConfirmation filesel_confirm_overwrite_cb
 	return ret;
 }
 #endif
+
+#ifdef G_OS_WIN32
+static gboolean is_ext_lnk(const gchar *filename)
+{
+	const gchar *ext;
+
+	if (filename && (ext = strrchr(filename, '.')) &&
+	    g_ascii_strcasecmp(ext, ".lnk") == 0)
+		return TRUE;
+
+	return FALSE;
+}
+
+static gchar *filesel_get_link(const gchar *link_file)
+{
+	WIN32_FIND_DATAW wfd;
+	IShellLinkW *psl;
+	IPersistFile *ppf;
+	HRESULT hr;
+	wchar_t *wlink_file;
+	wchar_t wtarget[MAX_PATH];
+	gchar *target = NULL;
+
+	wtarget[0] = 0L;
+
+	debug_print("link_file: %s\n", link_file);
+
+	CoInitialize(NULL);
+	if (S_OK == CoCreateInstance(&CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, &IID_IShellLinkW, (void **)&psl)) {
+		if (S_OK == IShellLinkW_QueryInterface(psl, &IID_IPersistFile, (void **)&ppf)) {
+			wlink_file = g_utf8_to_utf16(link_file, -1, NULL, NULL, NULL);
+			if (S_OK == IPersistFile_Load(ppf, wlink_file, STGM_READ)) {
+				if (S_OK == IShellLinkW_GetPath(psl, wtarget, MAX_PATH, &wfd, SLGP_UNCPRIORITY)) {
+					target = g_utf16_to_utf8(wtarget, -1, NULL, NULL, NULL);
+				}
+			}
+			IPersistFile_Release(ppf);
+			g_free(wlink_file);
+		}
+		IShellLinkW_Release(psl);
+	}
+	CoUninitialize();
+
+	if (target)
+		debug_print("target: %s\n", target);
+	else
+		debug_print("target not found\n");
+
+	return target;
+}
+
+static GSList *filesel_resolve_link(GtkFileChooser *chooser, GSList *list, gboolean *dir_selected)
+{
+	GSList *cur;
+	GSList *new_list = NULL;
+	gchar *target;
+
+	for (cur = list; cur != NULL; cur = cur->next) {
+		gchar *selected = (gchar *)cur->data;
+
+		if (is_ext_lnk(selected)) {
+			target = filesel_get_link(selected);
+
+		    	if (is_dir_exist(target)) {
+				gtk_file_chooser_set_current_folder(chooser, target);
+				g_free(target);
+				slist_free_strings(new_list);
+				g_slist_free(new_list);
+		    		new_list = NULL;
+				*dir_selected = TRUE;
+		    		break;
+			} else if (is_file_exist(target)) {
+				new_list = g_slist_append(new_list, target);
+			} else {
+			    	g_free(target);
+			}
+		} else {
+			new_list = g_slist_append(new_list, g_strdup(selected));
+		}
+	}
+
+	slist_free_strings(list);
+	g_slist_free(list);
+
+	return new_list;
+}
+#endif
